# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Optional, Tuple

import numpy as np
from numpy.typing import NDArray
from pyxpg import (
    AllocType,
    Buffer,
    BufferUsageFlags,
    CommandBuffer,
    ComputePipeline,
    DescriptorPool,
    DescriptorSet,
    DescriptorSetBinding,
    DescriptorSetLayout,
    DescriptorType,
    Image,
    ImageLayout,
    MemoryBarrier,
    MemoryUsage,
    PipelineStageFlags,
    PushConstantsRange,
    Shader,
    Stage,
)

from .utils.descriptors import create_descriptor_pool_and_sets_ringbuffer
from .utils.gpu import div_ceil, get_min_max_and_required_subgroup_size, view_bytes
from .utils.ring_buffer import RingBuffer

if TYPE_CHECKING:
    from .renderer import Renderer

BLOCK_SIZE_X = 8
BLOCK_SIZE_Y = 8
BLOCK_SIZE_Z = 8
BLOCK_TOTAL_SIZE = BLOCK_SIZE_X * BLOCK_SIZE_Y * BLOCK_SIZE_Z

COMPACT_GROUP_SIZE = 128


@dataclass
class MarchingCubesPipelineInstance:
    check_constants: NDArray[Any]
    compact_constants: NDArray[Any]
    march_constants: NDArray[Any]

    check_descriptor_pool: DescriptorPool
    check_descriptor_sets: RingBuffer[DescriptorSet]
    compact_descriptor_pool: DescriptorPool
    compact_descriptor_sets: RingBuffer[DescriptorSet]
    march_descriptor_pool: DescriptorPool
    march_descriptor_sets: RingBuffer[DescriptorSet]

    valid_blocks_counter_buf: Buffer
    vertices_counter_buf: Buffer
    indices_counter_buf: Buffer
    vertices_counter_readback_buf: Optional[Buffer]
    indices_counter_readback_buf: Optional[Buffer]


@dataclass
class MarchingCubesResult:
    positions: Buffer
    normals: Buffer
    indices: Buffer


class MarchingCubesPipeline:
    def __init__(self, r: "Renderer"):
        min_subgroup_size, max_subgroup_size, required_subgroup_size = get_min_max_and_required_subgroup_size(
            r.device, Stage.COMPUTE, 32, BLOCK_TOTAL_SIZE
        )

        defines = [
            ("BLOCK_SIZE_X", str(BLOCK_SIZE_X)),
            ("BLOCK_SIZE_Y", str(BLOCK_SIZE_Y)),
            ("BLOCK_SIZE_Z", str(BLOCK_SIZE_Z)),
            ("MIN_WAVE_SIZE", str(min_subgroup_size)),
            ("MAX_WAVE_SIZE", str(max_subgroup_size)),
        ]

        # Check step
        self.check_constants_dtype = np.dtype(
            {
                "level": (np.float32, 0),
                "num_blocks_x": (np.uint32, 4),
                "num_blocks_y": (np.uint32, 8),
            }
        )  # type: ignore

        self.check_descriptor_set_layout = DescriptorSetLayout(
            r.device,
            [
                DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),  # 0 - in volume
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 1 - out blocks
            ],
        )

        self.check_shader = r.compile_builtin_shader("3d/marching_cubes/check_surface.slang", defines=defines)
        self.check_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.check_shader.code),
            descriptor_set_layouts=[self.check_descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.check_constants_dtype.itemsize)],
            required_subgroup_size=required_subgroup_size,
        )

        # Compact step
        self.compact_constants_dtype = np.dtype(
            {
                "num_blocks": (np.uint32, 0),
            }
        )  # type: ignore

        self.compact_descriptor_set_layout = DescriptorSetLayout(
            r.device,
            [
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 0 - in blocks
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 1 - out valid block count
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 2 - out compacted blocks
            ],
        )

        self.compact_shader = r.compile_builtin_shader("3d/marching_cubes/compact.slang")
        self.compact_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.compact_shader.code),
            descriptor_set_layouts=[self.compact_descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.compact_constants_dtype.itemsize)],
            required_subgroup_size=required_subgroup_size,
        )

        # Marching cubes step
        self.march_constants_dtype = np.dtype(
            {
                "level": (np.float32, 0),
                "num_blocks_x": (np.uint32, 4),
                "num_blocks_y": (np.uint32, 8),
                "invert_normals": (np.uint32, 12),
                "volume_spacing": (np.dtype((np.float32, 3)), 16),
                "max_vertices": (np.uint32, 28),
                "max_indices": (np.uint32, 32),
            }
        )  # type: ignore

        self.march_descriptor_set_layout = DescriptorSetLayout(
            r.device,
            [
                DescriptorSetBinding(1, DescriptorType.STORAGE_IMAGE),  # 0 - in volume
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 1 - in compacted blocks
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 2 - in tri table
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 3 - out positions
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 4 - out normals
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 5 - out indices
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 6 - out num vertices
                DescriptorSetBinding(1, DescriptorType.STORAGE_BUFFER),  # 7 - out num indices
            ],
        )

        self.march_shader = r.compile_builtin_shader("3d/marching_cubes/marching_cubes.slang", defines=defines)
        self.march_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.march_shader.code),
            descriptor_set_layouts=[self.march_descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.march_constants_dtype.itemsize)],
            required_subgroup_size=required_subgroup_size,
        )

        self.march_count_shader = r.compile_builtin_shader(
            "3d/marching_cubes/marching_cubes.slang", defines=[*defines, ("COUNT_ONLY", "")]
        )
        self.march_count_pipeline = ComputePipeline(
            r.device,
            Shader(r.device, self.march_count_shader.code),
            descriptor_set_layouts=[self.march_descriptor_set_layout],
            push_constants_ranges=[PushConstantsRange(self.march_constants_dtype.itemsize)],
            required_subgroup_size=required_subgroup_size,
        )

        self.tri_table = Buffer.from_data(
            r.device,
            view_bytes(TRIS_TABLE),
            BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST,
            AllocType.DEVICE,
            name="marching-cubes-tri-table",
        )

        self.sync_instance = self.alloc_instance(r, single_set=True)

    def alloc_instance(self, r: "Renderer", single_set: bool = False) -> MarchingCubesPipelineInstance:
        # Initialize constants
        check_constants = np.zeros((1,), self.check_constants_dtype)
        compact_constants = np.zeros((1,), self.compact_constants_dtype)
        march_constants = np.zeros((1,), self.march_constants_dtype)

        # Allocate descriptor pools and sets
        number_of_sets = r.num_frames_in_flight if not single_set else 1
        check_pool, check_descriptor_sets = create_descriptor_pool_and_sets_ringbuffer(
            r.device,
            self.check_descriptor_set_layout,
            number_of_sets,
            name="marching-cubes-check",
        )
        compact_pool, compact_descriptor_sets = create_descriptor_pool_and_sets_ringbuffer(
            r.device,
            self.compact_descriptor_set_layout,
            number_of_sets,
            name="marching-cubes-compact",
        )
        march_pool, march_descriptor_sets = create_descriptor_pool_and_sets_ringbuffer(
            r.device,
            self.march_descriptor_set_layout,
            number_of_sets,
            name="marching-cubes-march",
        )

        # Allocate counter buffers
        valid_blocks_counter_buf = Buffer.from_data(
            r.device,
            view_bytes(np.array([0, 1, 1], np.uint32)),  # Alloc 3 elements to allow using this for indirect dispatch
            BufferUsageFlags.STORAGE
            | BufferUsageFlags.TRANSFER_DST
            | BufferUsageFlags.INDIRECT
            | BufferUsageFlags.TRANSFER_SRC,
            AllocType.DEVICE,
            name="marching-cubes-valid-blocks-counter",
        )

        vertices_counter_buf = Buffer.from_data(
            r.device,
            view_bytes(np.zeros((1,), np.uint32)),
            BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_DST | BufferUsageFlags.TRANSFER_SRC,
            AllocType.DEVICE_MAPPED_WITH_FALLBACK,
            name="marching-cubes-vertices-counter",
        )
        if not vertices_counter_buf.is_mapped:
            vertices_counter_readback_buf = Buffer.from_data(
                r.device,
                view_bytes(np.zeros((1,), np.uint32)),
                BufferUsageFlags.TRANSFER_DST,
                AllocType.HOST,
                name="marching-cubes-vertices-counter-readback",
            )
        else:
            vertices_counter_readback_buf = None

        # Allocate additional elements for use as draw indexed indirect argument
        indices_counter_buf = Buffer.from_data(
            r.device,
            view_bytes(np.array([0, 1, 0, 0, 0], np.uint32)),
            BufferUsageFlags.STORAGE
            | BufferUsageFlags.TRANSFER_DST
            | BufferUsageFlags.TRANSFER_SRC
            | BufferUsageFlags.INDIRECT,
            AllocType.DEVICE_MAPPED_WITH_FALLBACK,
            name="marching-cubes-indices-counter",
        )
        if not indices_counter_buf.is_mapped:
            indices_counter_readback_buf = Buffer.from_data(
                r.device,
                view_bytes(np.zeros((1,), np.uint32)),
                BufferUsageFlags.TRANSFER_DST,
                AllocType.HOST,
                name="marching-cubes-indices-counter-readback",
            )
        else:
            indices_counter_readback_buf = None

        # Write static descriptors
        for s in compact_descriptor_sets:
            s.write_buffer(valid_blocks_counter_buf, DescriptorType.STORAGE_BUFFER, 1)

        for s in march_descriptor_sets:
            s.write_buffer(self.tri_table, DescriptorType.STORAGE_BUFFER, 2)
            s.write_buffer(vertices_counter_buf, DescriptorType.STORAGE_BUFFER, 6)
            s.write_buffer(indices_counter_buf, DescriptorType.STORAGE_BUFFER, 7)

        return MarchingCubesPipelineInstance(
            check_constants,
            compact_constants,
            march_constants,
            check_pool,
            check_descriptor_sets,
            compact_pool,
            compact_descriptor_sets,
            march_pool,
            march_descriptor_sets,
            valid_blocks_counter_buf,
            vertices_counter_buf,
            indices_counter_buf,
            vertices_counter_readback_buf,
            indices_counter_readback_buf,
        )

    def run(
        self,
        cmd: CommandBuffer,
        width: int,
        height: int,
        depth: int,
        blocks_x: int,
        blocks_y: int,
        blocks_z: int,
        total_blocks: int,
        max_vertices: int,
        max_indices: int,
        instance: MarchingCubesPipelineInstance,
        size: Tuple[float, float, float],
        level: float = 0.0,
        invert_normals: bool = False,
    ) -> None:
        instance.check_constants["level"] = level
        instance.check_constants["num_blocks_x"] = blocks_x
        instance.check_constants["num_blocks_y"] = blocks_y

        instance.compact_constants["num_blocks"] = total_blocks

        instance.march_constants["level"] = level
        instance.march_constants["num_blocks_x"] = blocks_x
        instance.march_constants["num_blocks_y"] = blocks_y
        instance.march_constants["invert_normals"] = invert_normals
        instance.march_constants["volume_spacing"] = (size[0] / width, size[1] / height, size[2] / depth)
        instance.march_constants["max_vertices"] = max_vertices
        instance.march_constants["max_indices"] = max_indices

        check_set = instance.check_descriptor_sets.get_current_and_advance()
        compact_set = instance.compact_descriptor_sets.get_current_and_advance()
        march_set = instance.march_descriptor_sets.get_current_and_advance()

        cmd.bind_compute_pipeline(
            self.check_pipeline, descriptor_sets=[check_set], push_constants=instance.check_constants.tobytes()
        )
        cmd.dispatch(blocks_x, blocks_y, blocks_z)

        cmd.fill_buffer(instance.valid_blocks_counter_buf, 0, 4)

        cmd.memory_barrier_full(
            MemoryBarrier(
                src_stage=PipelineStageFlags.COMPUTE_SHADER | PipelineStageFlags.TRANSFER,
                dst_stage=PipelineStageFlags.COMPUTE_SHADER,
            )
        )

        cmd.bind_compute_pipeline(
            self.compact_pipeline,
            descriptor_sets=[compact_set],
            push_constants=instance.compact_constants.tobytes(),
        )
        cmd.dispatch(div_ceil(total_blocks, COMPACT_GROUP_SIZE), 1, 1)

        cmd.fill_buffer(instance.vertices_counter_buf, 0, 4)
        cmd.fill_buffer(instance.indices_counter_buf, 0, 4)

        cmd.memory_barrier_full(
            MemoryBarrier(
                src_stage=PipelineStageFlags.COMPUTE_SHADER | PipelineStageFlags.TRANSFER,
                dst_stage=PipelineStageFlags.COMPUTE_SHADER | PipelineStageFlags.DRAW_INDIRECT,
            )
        )

        cmd.bind_compute_pipeline(
            self.march_pipeline,
            descriptor_sets=[march_set],
            push_constants=instance.march_constants.tobytes(),
        )
        cmd.dispatch_indirect(instance.valid_blocks_counter_buf, 0)

        # Prepare for readback of vertices and indices count
        readback_src_stage = PipelineStageFlags.NONE
        if instance.vertices_counter_readback_buf is None or instance.indices_counter_readback_buf is None:
            readback_src_stage |= PipelineStageFlags.COMPUTE_SHADER
        if instance.vertices_counter_readback_buf is not None or instance.indices_counter_readback_buf is not None:
            # Ensure compute is done before issuing copy for readback
            readback_src_stage |= PipelineStageFlags.TRANSFER
            cmd.memory_barrier_full(
                MemoryBarrier(src_stage=PipelineStageFlags.COMPUTE_SHADER, dst_stage=PipelineStageFlags.TRANSFER)
            )
            if instance.vertices_counter_readback_buf is not None:
                cmd.copy_buffer(instance.vertices_counter_buf, instance.vertices_counter_readback_buf)
            if instance.indices_counter_readback_buf is not None:
                cmd.copy_buffer_range(instance.indices_counter_buf, instance.indices_counter_readback_buf, 4)

        cmd.barriers(
            memory_barriers=[
                MemoryBarrier(
                    src_stage=PipelineStageFlags.COMPUTE_SHADER,
                    dst_stage=PipelineStageFlags.DRAW_INDIRECT
                    | PipelineStageFlags.VERTEX_INPUT
                    | PipelineStageFlags.INDEX_INPUT,
                ),
                MemoryBarrier(
                    src_stage=readback_src_stage,
                    dst_stage=PipelineStageFlags.HOST,
                ),
            ]
        )

    def run_sync(
        self,
        r: "Renderer",
        sdf: Image,
        size: Tuple[float, float, float],
        level: float = 0.0,
        invert_normals: bool = False,
    ) -> MarchingCubesResult:
        width, height, depth = sdf.width, sdf.height, sdf.depth
        blocks_x = div_ceil(width, BLOCK_SIZE_X - 1)
        blocks_y = div_ceil(height, BLOCK_SIZE_Y - 1)
        blocks_z = div_ceil(depth, BLOCK_SIZE_Z - 1)
        total_blocks = blocks_x * blocks_y * blocks_z

        blocks_buffer = Buffer(
            r.device,
            total_blocks * 4,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC),
            AllocType.DEVICE,
            name="marching-cubes-sync-blocks",
        )
        valid_blocks_buffer = Buffer(
            r.device,
            total_blocks * 4,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC),
            AllocType.DEVICE,
            name="marching-cubes-sync-valid-blocks",
        )

        instance = self.sync_instance
        check_set = instance.check_descriptor_sets.get_current()
        check_set.write_image(sdf, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0)
        check_set.write_buffer(blocks_buffer, DescriptorType.STORAGE_BUFFER, 1)

        compact_set = instance.compact_descriptor_sets.get_current()
        compact_set.write_buffer(blocks_buffer, DescriptorType.STORAGE_BUFFER, 0)
        compact_set.write_buffer(valid_blocks_buffer, DescriptorType.STORAGE_BUFFER, 2)

        march_set = instance.march_descriptor_sets.get_current()
        march_set.write_image(sdf, ImageLayout.GENERAL, DescriptorType.STORAGE_IMAGE, 0)
        march_set.write_buffer(valid_blocks_buffer, DescriptorType.STORAGE_BUFFER, 1)
        march_set.write_buffer(r.zero_buffer, DescriptorType.STORAGE_BUFFER, 3)
        march_set.write_buffer(r.zero_buffer, DescriptorType.STORAGE_BUFFER, 4)
        march_set.write_buffer(r.zero_buffer, DescriptorType.STORAGE_BUFFER, 5)

        instance.check_constants["level"] = level
        instance.check_constants["num_blocks_x"] = blocks_x
        instance.check_constants["num_blocks_y"] = blocks_y

        instance.compact_constants["num_blocks"] = total_blocks

        instance.march_constants["level"] = level
        instance.march_constants["num_blocks_x"] = blocks_x
        instance.march_constants["num_blocks_y"] = blocks_y
        instance.march_constants["invert_normals"] = invert_normals
        instance.march_constants["volume_spacing"] = (size[0] / width, size[1] / height, size[2] / depth)
        instance.march_constants["max_vertices"] = 0
        instance.march_constants["max_indices"] = 0

        with r.device.sync_commands() as cmd:
            cmd.bind_compute_pipeline(
                self.check_pipeline, descriptor_sets=[check_set], push_constants=instance.check_constants.tobytes()
            )
            cmd.dispatch(blocks_x, blocks_y, blocks_z)

            cmd.fill_buffer(instance.valid_blocks_counter_buf, 0, 4)

            cmd.memory_barrier_full(
                MemoryBarrier(
                    src_stage=PipelineStageFlags.COMPUTE_SHADER | PipelineStageFlags.TRANSFER,
                    dst_stage=PipelineStageFlags.COMPUTE_SHADER,
                )
            )

            cmd.bind_compute_pipeline(
                self.compact_pipeline,
                descriptor_sets=[compact_set],
                push_constants=instance.compact_constants.tobytes(),
            )
            cmd.dispatch(div_ceil(total_blocks, COMPACT_GROUP_SIZE), 1, 1)

            cmd.memory_barrier_full(
                MemoryBarrier(
                    src_stage=PipelineStageFlags.COMPUTE_SHADER,
                    dst_stage=PipelineStageFlags.COMPUTE_SHADER | PipelineStageFlags.DRAW_INDIRECT,
                )
            )

            cmd.bind_compute_pipeline(
                self.march_count_pipeline,
                descriptor_sets=[march_set],
                push_constants=instance.march_constants.tobytes(),
            )
            cmd.dispatch_indirect(instance.valid_blocks_counter_buf, 0)

            readback_src_stage = PipelineStageFlags.NONE
            if instance.vertices_counter_readback_buf is None or instance.indices_counter_readback_buf is None:
                readback_src_stage |= PipelineStageFlags.COMPUTE_SHADER
            if instance.vertices_counter_readback_buf is not None or instance.indices_counter_readback_buf is not None:
                # Ensure compute is done before issuing copy for readback
                readback_src_stage |= PipelineStageFlags.TRANSFER
                cmd.memory_barrier_full(
                    MemoryBarrier(src_stage=PipelineStageFlags.COMPUTE_SHADER, dst_stage=PipelineStageFlags.TRANSFER)
                )

            # Copy to readback buffer
            if instance.vertices_counter_readback_buf is not None:
                cmd.copy_buffer(instance.vertices_counter_buf, instance.vertices_counter_readback_buf)
            if instance.indices_counter_readback_buf is not None:
                cmd.copy_buffer_range(instance.indices_counter_buf, instance.indices_counter_readback_buf, 4)

            # Ensure copy is done before clearing counter
            if instance.vertices_counter_readback_buf is not None or instance.indices_counter_readback_buf is not None:
                cmd.memory_barrier_full(
                    MemoryBarrier(src_stage=PipelineStageFlags.TRANSFER, dst_stage=PipelineStageFlags.TRANSFER)
                )

            # Clear GPU counter buffer. If the buffer is mapped we do this from the CPU later instead
            if instance.vertices_counter_readback_buf is not None:
                cmd.fill_buffer(instance.vertices_counter_buf, 0, 4)
            if instance.indices_counter_readback_buf is not None:
                cmd.fill_buffer(instance.indices_counter_buf, 0, 4)

            # Ensure copy is visible to host (either written form compute or from copy)
            cmd.memory_barrier_full(MemoryBarrier(src_stage=readback_src_stage, dst_stage=PipelineStageFlags.HOST))

        vertices_counter_readback_buf = (
            instance.vertices_counter_readback_buf
            if instance.vertices_counter_readback_buf is not None
            else instance.vertices_counter_buf
        )
        indices_counter_readback_buf = (
            instance.indices_counter_readback_buf
            if instance.indices_counter_readback_buf is not None
            else instance.indices_counter_buf
        )

        vertices_arr = np.frombuffer(vertices_counter_readback_buf.data, np.uint32)
        vertices_count = vertices_arr[0]
        vertices_arr[0] = 0

        indices_arr = np.frombuffer(indices_counter_readback_buf.data, np.uint32)
        indices_count = indices_arr[0]
        indices_arr[0] = 0

        positions_buf = Buffer(
            r.device,
            vertices_count * 12,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC | BufferUsageFlags.VERTEX),
            AllocType.DEVICE,
            name="marching-cubes-sync-positions",
        )
        normals_buf = Buffer(
            r.device,
            vertices_count * 12,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC | BufferUsageFlags.VERTEX),
            AllocType.DEVICE,
            name="marching-cubes-sync-normals",
        )
        indices_buf = Buffer(
            r.device,
            indices_count * 4,
            (BufferUsageFlags.STORAGE | BufferUsageFlags.TRANSFER_SRC | BufferUsageFlags.INDEX),
            AllocType.DEVICE,
            name="marching-cubes-sync-indices",
        )
        march_set.write_buffer(positions_buf, DescriptorType.STORAGE_BUFFER, 3)
        march_set.write_buffer(normals_buf, DescriptorType.STORAGE_BUFFER, 4)
        march_set.write_buffer(indices_buf, DescriptorType.STORAGE_BUFFER, 5)

        instance.march_constants["max_vertices"] = vertices_count
        instance.march_constants["max_indices"] = indices_count
        with r.device.sync_commands() as cmd:
            cmd.bind_compute_pipeline(
                self.march_pipeline, descriptor_sets=[march_set], push_constants=instance.march_constants.tobytes()
            )
            cmd.dispatch_indirect(instance.valid_blocks_counter_buf, 0)
            cmd.memory_barrier(MemoryUsage.COMPUTE_SHADER, MemoryUsage.ALL)

        blocks_buffer.destroy()
        valid_blocks_buffer.destroy()

        return MarchingCubesResult(positions_buf, normals_buf, indices_buf)


X = 255
TRIS_TABLE = np.array(
    [
        [X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [0, 8, 3, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [0, 1, 9, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [1, 8, 3, 9, 8, 1, X, X, X, X, X, X, X, X, X, X],
        [1, 2, 10, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [0, 8, 3, 1, 2, 10, X, X, X, X, X, X, X, X, X, X],
        [9, 2, 10, 0, 2, 9, X, X, X, X, X, X, X, X, X, X],
        [2, 8, 3, 2, 10, 8, 10, 9, 8, X, X, X, X, X, X, X],
        [3, 11, 2, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [0, 11, 2, 8, 11, 0, X, X, X, X, X, X, X, X, X, X],
        [1, 9, 0, 2, 3, 11, X, X, X, X, X, X, X, X, X, X],
        [1, 11, 2, 1, 9, 11, 9, 8, 11, X, X, X, X, X, X, X],
        [3, 10, 1, 11, 10, 3, X, X, X, X, X, X, X, X, X, X],
        [0, 10, 1, 0, 8, 10, 8, 11, 10, X, X, X, X, X, X, X],
        [3, 9, 0, 3, 11, 9, 11, 10, 9, X, X, X, X, X, X, X],
        [9, 8, 10, 10, 8, 11, X, X, X, X, X, X, X, X, X, X],
        [4, 7, 8, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [4, 3, 0, 7, 3, 4, X, X, X, X, X, X, X, X, X, X],
        [0, 1, 9, 8, 4, 7, X, X, X, X, X, X, X, X, X, X],
        [4, 1, 9, 4, 7, 1, 7, 3, 1, X, X, X, X, X, X, X],
        [1, 2, 10, 8, 4, 7, X, X, X, X, X, X, X, X, X, X],
        [3, 4, 7, 3, 0, 4, 1, 2, 10, X, X, X, X, X, X, X],
        [9, 2, 10, 9, 0, 2, 8, 4, 7, X, X, X, X, X, X, X],
        [2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, X, X, X, X],
        [8, 4, 7, 3, 11, 2, X, X, X, X, X, X, X, X, X, X],
        [11, 4, 7, 11, 2, 4, 2, 0, 4, X, X, X, X, X, X, X],
        [9, 0, 1, 8, 4, 7, 2, 3, 11, X, X, X, X, X, X, X],
        [4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, X, X, X, X],
        [3, 10, 1, 3, 11, 10, 7, 8, 4, X, X, X, X, X, X, X],
        [1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, X, X, X, X],
        [4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, X, X, X, X],
        [4, 7, 11, 4, 11, 9, 9, 11, 10, X, X, X, X, X, X, X],
        [9, 5, 4, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [9, 5, 4, 0, 8, 3, X, X, X, X, X, X, X, X, X, X],
        [0, 5, 4, 1, 5, 0, X, X, X, X, X, X, X, X, X, X],
        [8, 5, 4, 8, 3, 5, 3, 1, 5, X, X, X, X, X, X, X],
        [1, 2, 10, 9, 5, 4, X, X, X, X, X, X, X, X, X, X],
        [3, 0, 8, 1, 2, 10, 4, 9, 5, X, X, X, X, X, X, X],
        [5, 2, 10, 5, 4, 2, 4, 0, 2, X, X, X, X, X, X, X],
        [2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, X, X, X, X],
        [9, 5, 4, 2, 3, 11, X, X, X, X, X, X, X, X, X, X],
        [0, 11, 2, 0, 8, 11, 4, 9, 5, X, X, X, X, X, X, X],
        [0, 5, 4, 0, 1, 5, 2, 3, 11, X, X, X, X, X, X, X],
        [2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, X, X, X, X],
        [10, 3, 11, 10, 1, 3, 9, 5, 4, X, X, X, X, X, X, X],
        [4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, X, X, X, X],
        [5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, X, X, X, X],
        [5, 4, 8, 5, 8, 10, 10, 8, 11, X, X, X, X, X, X, X],
        [9, 7, 8, 5, 7, 9, X, X, X, X, X, X, X, X, X, X],
        [9, 3, 0, 9, 5, 3, 5, 7, 3, X, X, X, X, X, X, X],
        [0, 7, 8, 0, 1, 7, 1, 5, 7, X, X, X, X, X, X, X],
        [1, 5, 3, 3, 5, 7, X, X, X, X, X, X, X, X, X, X],
        [9, 7, 8, 9, 5, 7, 10, 1, 2, X, X, X, X, X, X, X],
        [10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, X, X, X, X],
        [8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, X, X, X, X],
        [2, 10, 5, 2, 5, 3, 3, 5, 7, X, X, X, X, X, X, X],
        [7, 9, 5, 7, 8, 9, 3, 11, 2, X, X, X, X, X, X, X],
        [9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, X, X, X, X],
        [2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, X, X, X, X],
        [11, 2, 1, 11, 1, 7, 7, 1, 5, X, X, X, X, X, X, X],
        [9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, X, X, X, X],
        [5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, X],
        [11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, X],
        [11, 10, 5, 7, 11, 5, X, X, X, X, X, X, X, X, X, X],
        [10, 6, 5, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [0, 8, 3, 5, 10, 6, X, X, X, X, X, X, X, X, X, X],
        [9, 0, 1, 5, 10, 6, X, X, X, X, X, X, X, X, X, X],
        [1, 8, 3, 1, 9, 8, 5, 10, 6, X, X, X, X, X, X, X],
        [1, 6, 5, 2, 6, 1, X, X, X, X, X, X, X, X, X, X],
        [1, 6, 5, 1, 2, 6, 3, 0, 8, X, X, X, X, X, X, X],
        [9, 6, 5, 9, 0, 6, 0, 2, 6, X, X, X, X, X, X, X],
        [5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, X, X, X, X],
        [2, 3, 11, 10, 6, 5, X, X, X, X, X, X, X, X, X, X],
        [11, 0, 8, 11, 2, 0, 10, 6, 5, X, X, X, X, X, X, X],
        [0, 1, 9, 2, 3, 11, 5, 10, 6, X, X, X, X, X, X, X],
        [5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, X, X, X, X],
        [6, 3, 11, 6, 5, 3, 5, 1, 3, X, X, X, X, X, X, X],
        [0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, X, X, X, X],
        [3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, X, X, X, X],
        [6, 5, 9, 6, 9, 11, 11, 9, 8, X, X, X, X, X, X, X],
        [5, 10, 6, 4, 7, 8, X, X, X, X, X, X, X, X, X, X],
        [4, 3, 0, 4, 7, 3, 6, 5, 10, X, X, X, X, X, X, X],
        [1, 9, 0, 5, 10, 6, 8, 4, 7, X, X, X, X, X, X, X],
        [10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, X, X, X, X],
        [6, 1, 2, 6, 5, 1, 4, 7, 8, X, X, X, X, X, X, X],
        [1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, X, X, X, X],
        [8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, X, X, X, X],
        [7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, X],
        [3, 11, 2, 7, 8, 4, 10, 6, 5, X, X, X, X, X, X, X],
        [5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, X, X, X, X],
        [0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, X, X, X, X],
        [9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, X],
        [8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, X, X, X, X],
        [5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, X],
        [0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, X],
        [6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, X, X, X, X],
        [10, 4, 9, 6, 4, 10, X, X, X, X, X, X, X, X, X, X],
        [4, 10, 6, 4, 9, 10, 0, 8, 3, X, X, X, X, X, X, X],
        [10, 0, 1, 10, 6, 0, 6, 4, 0, X, X, X, X, X, X, X],
        [8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, X, X, X, X],
        [1, 4, 9, 1, 2, 4, 2, 6, 4, X, X, X, X, X, X, X],
        [3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, X, X, X, X],
        [0, 2, 4, 4, 2, 6, X, X, X, X, X, X, X, X, X, X],
        [8, 3, 2, 8, 2, 4, 4, 2, 6, X, X, X, X, X, X, X],
        [10, 4, 9, 10, 6, 4, 11, 2, 3, X, X, X, X, X, X, X],
        [0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, X, X, X, X],
        [3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, X, X, X, X],
        [6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, X],
        [9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, X, X, X, X],
        [8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, X],
        [3, 11, 6, 3, 6, 0, 0, 6, 4, X, X, X, X, X, X, X],
        [6, 4, 8, 11, 6, 8, X, X, X, X, X, X, X, X, X, X],
        [7, 10, 6, 7, 8, 10, 8, 9, 10, X, X, X, X, X, X, X],
        [0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, X, X, X, X],
        [10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, X, X, X, X],
        [10, 6, 7, 10, 7, 1, 1, 7, 3, X, X, X, X, X, X, X],
        [1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, X, X, X, X],
        [2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, X],
        [7, 8, 0, 7, 0, 6, 6, 0, 2, X, X, X, X, X, X, X],
        [7, 3, 2, 6, 7, 2, X, X, X, X, X, X, X, X, X, X],
        [2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, X, X, X, X],
        [2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, X],
        [1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, X],
        [11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, X, X, X, X],
        [8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, X],
        [0, 9, 1, 11, 6, 7, X, X, X, X, X, X, X, X, X, X],
        [7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, X, X, X, X],
        [7, 11, 6, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [7, 6, 11, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [3, 0, 8, 11, 7, 6, X, X, X, X, X, X, X, X, X, X],
        [0, 1, 9, 11, 7, 6, X, X, X, X, X, X, X, X, X, X],
        [8, 1, 9, 8, 3, 1, 11, 7, 6, X, X, X, X, X, X, X],
        [10, 1, 2, 6, 11, 7, X, X, X, X, X, X, X, X, X, X],
        [1, 2, 10, 3, 0, 8, 6, 11, 7, X, X, X, X, X, X, X],
        [2, 9, 0, 2, 10, 9, 6, 11, 7, X, X, X, X, X, X, X],
        [6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, X, X, X, X],
        [7, 2, 3, 6, 2, 7, X, X, X, X, X, X, X, X, X, X],
        [7, 0, 8, 7, 6, 0, 6, 2, 0, X, X, X, X, X, X, X],
        [2, 7, 6, 2, 3, 7, 0, 1, 9, X, X, X, X, X, X, X],
        [1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, X, X, X, X],
        [10, 7, 6, 10, 1, 7, 1, 3, 7, X, X, X, X, X, X, X],
        [10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, X, X, X, X],
        [0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, X, X, X, X],
        [7, 6, 10, 7, 10, 8, 8, 10, 9, X, X, X, X, X, X, X],
        [6, 8, 4, 11, 8, 6, X, X, X, X, X, X, X, X, X, X],
        [3, 6, 11, 3, 0, 6, 0, 4, 6, X, X, X, X, X, X, X],
        [8, 6, 11, 8, 4, 6, 9, 0, 1, X, X, X, X, X, X, X],
        [9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, X, X, X, X],
        [6, 8, 4, 6, 11, 8, 2, 10, 1, X, X, X, X, X, X, X],
        [1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, X, X, X, X],
        [4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, X, X, X, X],
        [10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, X],
        [8, 2, 3, 8, 4, 2, 4, 6, 2, X, X, X, X, X, X, X],
        [0, 4, 2, 4, 6, 2, X, X, X, X, X, X, X, X, X, X],
        [1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, X, X, X, X],
        [1, 9, 4, 1, 4, 2, 2, 4, 6, X, X, X, X, X, X, X],
        [8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, X, X, X, X],
        [10, 1, 0, 10, 0, 6, 6, 0, 4, X, X, X, X, X, X, X],
        [4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, X],
        [10, 9, 4, 6, 10, 4, X, X, X, X, X, X, X, X, X, X],
        [4, 9, 5, 7, 6, 11, X, X, X, X, X, X, X, X, X, X],
        [0, 8, 3, 4, 9, 5, 11, 7, 6, X, X, X, X, X, X, X],
        [5, 0, 1, 5, 4, 0, 7, 6, 11, X, X, X, X, X, X, X],
        [11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, X, X, X, X],
        [9, 5, 4, 10, 1, 2, 7, 6, 11, X, X, X, X, X, X, X],
        [6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, X, X, X, X],
        [7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, X, X, X, X],
        [3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, X],
        [7, 2, 3, 7, 6, 2, 5, 4, 9, X, X, X, X, X, X, X],
        [9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, X, X, X, X],
        [3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, X, X, X, X],
        [6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, X],
        [9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, X, X, X, X],
        [1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, X],
        [4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, X],
        [7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, X, X, X, X],
        [6, 9, 5, 6, 11, 9, 11, 8, 9, X, X, X, X, X, X, X],
        [3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, X, X, X, X],
        [0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, X, X, X, X],
        [6, 11, 3, 6, 3, 5, 5, 3, 1, X, X, X, X, X, X, X],
        [1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, X, X, X, X],
        [0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, X],
        [11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, X],
        [6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, X, X, X, X],
        [5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, X, X, X, X],
        [9, 5, 6, 9, 6, 0, 0, 6, 2, X, X, X, X, X, X, X],
        [1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, X],
        [1, 5, 6, 2, 1, 6, X, X, X, X, X, X, X, X, X, X],
        [1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, X],
        [10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, X, X, X, X],
        [0, 3, 8, 5, 6, 10, X, X, X, X, X, X, X, X, X, X],
        [10, 5, 6, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [11, 5, 10, 7, 5, 11, X, X, X, X, X, X, X, X, X, X],
        [11, 5, 10, 11, 7, 5, 8, 3, 0, X, X, X, X, X, X, X],
        [5, 11, 7, 5, 10, 11, 1, 9, 0, X, X, X, X, X, X, X],
        [10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, X, X, X, X],
        [11, 1, 2, 11, 7, 1, 7, 5, 1, X, X, X, X, X, X, X],
        [0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, X, X, X, X],
        [9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, X, X, X, X],
        [7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, X],
        [2, 5, 10, 2, 3, 5, 3, 7, 5, X, X, X, X, X, X, X],
        [8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, X, X, X, X],
        [9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, X, X, X, X],
        [9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, X],
        [1, 3, 5, 3, 7, 5, X, X, X, X, X, X, X, X, X, X],
        [0, 8, 7, 0, 7, 1, 1, 7, 5, X, X, X, X, X, X, X],
        [9, 0, 3, 9, 3, 5, 5, 3, 7, X, X, X, X, X, X, X],
        [9, 8, 7, 5, 9, 7, X, X, X, X, X, X, X, X, X, X],
        [5, 8, 4, 5, 10, 8, 10, 11, 8, X, X, X, X, X, X, X],
        [5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, X, X, X, X],
        [0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, X, X, X, X],
        [10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, X],
        [2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, X, X, X, X],
        [0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, X],
        [0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, X],
        [9, 4, 5, 2, 11, 3, X, X, X, X, X, X, X, X, X, X],
        [2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, X, X, X, X],
        [5, 10, 2, 5, 2, 4, 4, 2, 0, X, X, X, X, X, X, X],
        [3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, X],
        [5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, X, X, X, X],
        [8, 4, 5, 8, 5, 3, 3, 5, 1, X, X, X, X, X, X, X],
        [0, 4, 5, 1, 0, 5, X, X, X, X, X, X, X, X, X, X],
        [8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, X, X, X, X],
        [9, 4, 5, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [4, 11, 7, 4, 9, 11, 9, 10, 11, X, X, X, X, X, X, X],
        [0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, X, X, X, X],
        [1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, X, X, X, X],
        [3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, X],
        [4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, X, X, X, X],
        [9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, X],
        [11, 7, 4, 11, 4, 2, 2, 4, 0, X, X, X, X, X, X, X],
        [11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, X, X, X, X],
        [2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, X, X, X, X],
        [9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, X],
        [3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, X],
        [1, 10, 2, 8, 7, 4, X, X, X, X, X, X, X, X, X, X],
        [4, 9, 1, 4, 1, 7, 7, 1, 3, X, X, X, X, X, X, X],
        [4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, X, X, X, X],
        [4, 0, 3, 7, 4, 3, X, X, X, X, X, X, X, X, X, X],
        [4, 8, 7, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [9, 10, 8, 10, 11, 8, X, X, X, X, X, X, X, X, X, X],
        [3, 0, 9, 3, 9, 11, 11, 9, 10, X, X, X, X, X, X, X],
        [0, 1, 10, 0, 10, 8, 8, 10, 11, X, X, X, X, X, X, X],
        [3, 1, 10, 11, 3, 10, X, X, X, X, X, X, X, X, X, X],
        [1, 2, 11, 1, 11, 9, 9, 11, 8, X, X, X, X, X, X, X],
        [3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, X, X, X, X],
        [0, 2, 11, 8, 0, 11, X, X, X, X, X, X, X, X, X, X],
        [3, 2, 11, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [2, 3, 8, 2, 8, 10, 10, 8, 9, X, X, X, X, X, X, X],
        [9, 10, 2, 0, 9, 2, X, X, X, X, X, X, X, X, X, X],
        [2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, X, X, X, X],
        [1, 10, 2, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [1, 3, 8, 9, 1, 8, X, X, X, X, X, X, X, X, X, X],
        [0, 9, 1, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [0, 3, 8, X, X, X, X, X, X, X, X, X, X, X, X, X],
        [X, X, X, X, X, X, X, X, X, X, X, X, X, X, X, X],
    ],
    dtype=np.uint32,
)



import os
import sys
from typing import Optional

import cv2
import numpy as np
from pyglm.glm import mat4x3, normalize, quatLookAtRH, vec3
from pyxpg import *

from ambra.config import CameraConfig, Config, GuiConfig, PlaybackConfig
from ambra.geometry import create_axis3d_lines_and_colors, create_cube
from ambra.lights import DirectionalLight, DirectionalShadowSettings, EnvironmentCubemaps
from ambra.primitives3d import Lines
from ambra.renderer import Renderer
from ambra.renderer_frame import RendererFrame
from ambra.scene import BufferProperty, Object3D
from ambra.utils.descriptors import create_descriptor_layout_pool_and_sets_ringbuffer
from ambra.viewer import Viewer
from ambra.viewport import Viewport

os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"


class DebugCube(Object3D):
    def __init__(
        self,
        cubemap: Image,
        positions: BufferProperty,
        indices: Optional[BufferProperty] = None,
        name: Optional[str] = None,
    ):
        self.constants_dtype = np.dtype(
            {
                "transform": (np.dtype((np.float32, (3, 4))), 0),
                "level": (np.uint32, 48),
            }
        )  # type: ignore
        self.constants = np.zeros((1,), self.constants_dtype)
        self.constants["transform"] = mat4x3(1.0)
        self.constants["level"] = 1

        super().__init__(name)

        self.positions = self.add_buffer_property(positions, np.float32, (-1, 3), name="positions").use_gpu(
            BufferUsageFlags.VERTEX, PipelineStageFlags.VERTEX_INPUT
        )
        self.indices = (
            self.add_buffer_property(indices, np.uint32, (-1,), name="indices").use_gpu(
                BufferUsageFlags.INDEX, PipelineStageFlags.VERTEX_INPUT
            )
            if indices is not None
            else None
        )
        self.cubemap = cubemap

    def create(self, r: Renderer) -> None:
        vertex_bindings = [
            VertexBinding(0, 12, VertexInputRate.VERTEX),
        ]
        vertex_attributes = [
            VertexAttribute(0, 0, Format.R32G32B32_SFLOAT),
        ]

        self.descriptor_set_layout, self.descriptor_pool, self.descriptor_sets = (
            create_descriptor_layout_pool_and_sets_ringbuffer(
                r.device,
                [
                    DescriptorSetBinding(1, DescriptorType.COMBINED_IMAGE_SAMPLER),
                ],
                r.num_frames_in_flight,
            )
        )
        self.sampler = Sampler(r.device, Filter.NEAREST, Filter.NEAREST)
        # self.sampler = Sampler(r.device, Filter.LINEAR, Filter.LINEAR)

        vert = r.compile_builtin_shader("3d/cube.slang", "vertex_main")
        frag = r.compile_builtin_shader("3d/cube.slang", "pixel_main")

        self.pipeline = GraphicsPipeline(
            r.device,
            stages=[
                PipelineStage(Shader(r.device, vert.code), Stage.VERTEX),
                PipelineStage(Shader(r.device, frag.code), Stage.FRAGMENT),
            ],
            vertex_bindings=vertex_bindings,
            vertex_attributes=vertex_attributes,
            input_assembly=InputAssembly(PrimitiveTopology.TRIANGLE_LIST),
            attachments=[Attachment(format=r.output_format)],
            depth=Depth(r.depth_format, True, True, r.depth_compare_op),
            descriptor_set_layouts=[
                r.scene_descriptor_set_layout,
                self.descriptor_set_layout,
            ],
            push_constants_ranges=[PushConstantsRange(self.constants_dtype.itemsize)],
        )

    def render(
        self, r: Renderer, frame: RendererFrame, viewport: Viewport, scene_descriptor_set: DescriptorSet
    ) -> None:
        vertex_buffers = [
            self.positions.get_current_gpu().buffer_and_offset(),
        ]

        self.constants["level"] = level

        descriptor_set = self.descriptor_sets.get_current_and_advance()
        descriptor_set.write_combined_image_sampler(
            self.cubemap, ImageLayout.SHADER_READ_ONLY_OPTIMAL, self.sampler, 0
        )

        frame.cmd.bind_graphics_pipeline(
            self.pipeline,
            vertex_buffers=vertex_buffers,
            index_buffer=self.indices.get_current_gpu().buffer_and_offset() if self.indices is not None else None,
            descriptor_sets=[
                scene_descriptor_set,
                descriptor_set,
            ],
            push_constants=self.constants.tobytes(),
        )

        if self.indices is not None:
            frame.cmd.draw_indexed(self.indices.get_current().shape[0])
        else:
            frame.cmd.draw(self.positions.get_current().shape[0])


level = 0


def main():
    class CustomViewer(Viewer):
        def on_key(self, key: Key, action: Action, modifiers: Modifiers):
            global level
            if action == Action.PRESS:
                if key == Key.Z:
                    level = max(level - 1, 0)
                if key == Key.X:
                    level = min(level + 1, 10)
                if key == Key.V:
                    cube.cubemap = result.irradiance_cubemap
                if key == Key.B:
                    cube.cubemap = result.specular_cubemap
            return super().on_key(key, action, modifiers)

    viewer = CustomViewer(
        config=Config(
            playback=PlaybackConfig(
                playing=True,
            ),
            gui=GuiConfig(
                stats=True,
            ),
            world_up=(0, 0, 1),
            camera=CameraConfig(
                position=(3, 3, 3),
                target=(0, 0, 0),
            ),
        ),
    )

    device = viewer.device

    # Load image
    hdr_data = cv2.imread(sys.argv[1], cv2.IMREAD_ANYCOLOR | cv2.IMREAD_ANYDEPTH | cv2.IMREAD_COLOR_RGB)
    hdr_height, hdr_width, hdr_channels = hdr_data.shape
    assert hdr_channels == 3 and hdr_data.dtype == np.float32, (
        f"Format not supported {hdr_channels} channels {hdr_data.dtype} type"
    )
    hdr_data = np.dstack((hdr_data, np.ones((hdr_height, hdr_width, 1), np.float32)))
    hdr_img = Image.from_data(
        device,
        hdr_data,
        ImageLayout.SHADER_READ_ONLY_OPTIMAL,
        hdr_width,
        hdr_height,
        Format.R32G32B32A32_SFLOAT,
        ImageUsageFlags.SAMPLED | ImageUsageFlags.TRANSFER_DST,
        AllocType.DEVICE,
    )

    result = viewer.renderer.run_ibl_pipeline(hdr_img)

    # Test round trip from GPU to CPU to disk and back.
    cubemap_data = result.cpu(device)
    file = "cubemap_ibl.npz"
    cubemap_data.save(file)
    cubemap_data = EnvironmentCubemaps.load(file)
    result = cubemap_data.gpu(device)

    cube_positions, _, cube_faces = create_cube(extents=(2.0, 2.0, 2.0))
    cube = DebugCube(result.specular_cubemap, cube_positions, cube_faces)

    positions, colors = create_axis3d_lines_and_colors()
    line = Lines(positions * 3, colors, 4, translation=(-1, -1, -1))

    light_position = vec3(5, 6, 7) * 4.0
    light_target = vec3(0, 0, 0)
    rotation = quatLookAtRH(normalize(light_target - light_position), vec3(0, 0, 1))
    light = DirectionalLight(
        np.array([1.0, 1.0, 1.0]),
        shadow_settings=DirectionalShadowSettings(half_extent=10.0, z_near=1.0, z_far=100),
        translation=light_position,
        rotation=rotation,
    )

    viewer.scene.objects.extend([cube, line, light])

    viewer.run()


if __name__ == "__main__":
    main()



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

from typing import Generic, Iterator, List, TypeVar

R = TypeVar("R")


class RingBuffer(Generic[R]):
    def __init__(self, items: List[R]):
        self.items = items
        self.index = 0

    def __len__(self) -> int:
        return self.items.__len__()

    def __iter__(self) -> Iterator[R]:
        return self.items.__iter__()

    def __repr__(self) -> str:
        return self.items.__repr__()

    def get_current(self) -> R:
        return self.items[self.index]

    def advance(self) -> None:
        self.index = (self.index + 1) % len(self.items)

    def set(self, index: int) -> None:
        self.index = index % len(self.items)

    def get_current_and_advance(self) -> R:
        cur = self.get_current()
        self.advance()
        return cur



# Copyright Dario Mylonopoulos
# SPDX-License-Identifier: MIT

import bisect
from dataclasses import dataclass
from enum import Enum, auto
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Tuple, Union

import numpy as np
from numpy.typing import ArrayLike, DTypeLike, NDArray
from pyxpg import (
    BufferUsageFlags,
    DeviceFeatures,
    Format,
    ImageLayout,
    ImageUsageFlags,
    PipelineStageFlags,
)

from . import gpu_property as gpu_property_mod
from .utils.gpu import format_from_channels_dtype, view_bytes

if TYPE_CHECKING:
    from .renderer import Renderer


PropertyItem = NDArray[Any]


class AnimationBoundary(Enum):
    HOLD = auto()
    REPEAT = auto()
    MIRROR = auto()
    DISABLE = auto()


class Animation:
    def __init__(self, boundary: AnimationBoundary = AnimationBoundary.HOLD):
        self.boundary = boundary

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        return -1

    def is_dynamic(self, n: int) -> bool:
        return False

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        return True

    def start_animation_time(self, n: int, fps: float) -> float:
        return 0

    def end_animation_time(self, n: int, fps: float) -> float:
        return 0

    def is_frame_index_disabled(self, frame: int, begin: int, end: int) -> bool:
        return (frame < begin) or (self.boundary == AnimationBoundary.DISABLE and frame >= end)

    def is_time_disabled(self, t: float, t_min: float, t_max: float) -> bool:
        return (t < t_min) or (self.boundary == AnimationBoundary.DISABLE and t >= t_max)

    def map_frame_index(self, frame: int, begin: int, end: int) -> int:
        if self.boundary == AnimationBoundary.HOLD or self.boundary == AnimationBoundary.DISABLE:
            return int(np.clip(frame, begin, end - 1))

        n = end - begin
        norm_frame = frame - begin
        if self.boundary == AnimationBoundary.REPEAT:
            return (norm_frame % n) + begin
        elif self.boundary == AnimationBoundary.MIRROR:
            dr = norm_frame % (2 * n)
            if dr >= n:
                return 2 * n - dr - 1 + begin
            else:
                return dr + begin
        else:
            raise ValueError(f"Unhandled enum variant: {self.boundary}")

    def map_time(self, t: float, t_min: float, t_max: float) -> float:
        if self.boundary == AnimationBoundary.HOLD or self.boundary == AnimationBoundary.DISABLE:
            return np.clip(t, t_min, t_max)  # type: ignore
        elif self.boundary == AnimationBoundary.REPEAT:
            return np.fmod(t - t_min, t_max - t_min) + t_min  # type: ignore
        elif self.boundary == AnimationBoundary.MIRROR:
            double_dt = (t_max - t_min) * 2.0
            dr: float = np.fmod(t - t_min, double_dt)
            if dr >= t_max - t_min:
                return double_dt - dr + t_min
            else:
                return dr + t_min
        else:
            raise ValueError(f"Unhandled enum variant: {self.boundary}")


class TimeSampledAnimation(Animation):
    def __init__(self, timestamps: NDArray[np.float64], boundary: AnimationBoundary = AnimationBoundary.HOLD):
        super().__init__(boundary)
        self.timestamps = np.asarray(timestamps)

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        count = self.timestamps.size
        if count == 0:
            return -1
        elif count == 1:
            return 0
        else:
            t = self.map_time(time, self.timestamps[0], self.timestamps[-1])
            return max(int(np.searchsorted(self.timestamps, t, side="right")) - 1, 0)

    def is_dynamic(self, n: int) -> bool:
        return True

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        count = self.timestamps.size
        if count == 0:
            return True
        else:
            return self.is_time_disabled(time, self.timestamps[0], self.timestamps[-1])

    def start_animation_time(self, n: int, fps: float) -> float:
        return self.timestamps[0] if self.timestamps else 0.0

    def end_animation_time(self, n: int, fps: float) -> float:
        return self.timestamps[-1] if self.timestamps.size > 0 else 0.0


class ListTimeSampledAnimation(Animation):
    def __init__(self, timestamps: List[float], boundary: AnimationBoundary = AnimationBoundary.HOLD):
        super().__init__(boundary)
        self.timestamps = timestamps

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        count = len(self.timestamps)
        if count == 0:
            return -1
        elif count == 1:
            return 0
        else:
            t = self.map_time(time, self.timestamps[0], self.timestamps[-1])
            return max(bisect.bisect_right(self.timestamps, t) - 1, 0)

    def is_dynamic(self, n: int) -> bool:
        return True

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        if not self.timestamps:
            return True
        else:
            return self.is_time_disabled(time, self.timestamps[0], self.timestamps[-1])

    def start_animation_time(self, n: int, fps: float) -> float:
        return self.timestamps[0] if self.timestamps else 0.0

    def end_animation_time(self, n: int, fps: float) -> float:
        return self.timestamps[-1] if self.timestamps else 0.0


class FrameSampledAnimation(Animation):
    def __init__(self, indices: NDArray[np.uint32], boundary: AnimationBoundary = AnimationBoundary.HOLD):
        super().__init__(boundary)
        self.indices = np.asarray(indices, np.uint32)

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        count = self.indices.size
        if count == 0:
            return -1
        elif count == 1:
            return 0
        else:
            idx = self.map_frame_index(frame_index, self.indices[0], self.indices[-1] + 1)
            return max(int(np.searchsorted(self.indices, idx, side="right")) - 1, 0)

    def is_dynamic(self, n: int) -> bool:
        return True

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        count = self.indices.size
        if count == 0:
            return True
        else:
            return self.is_frame_index_disabled(frame_index, self.indices[0], self.indices[-1])

    def start_animation_time(self, n: int, fps: float) -> float:
        return self.indices[0] / fps if self.indices else 0.0

    def end_animation_time(self, n: int, fps: float) -> float:
        return (self.indices[-1] + 1) / fps if self.indices.size > 0 else 0.0


class ListFrameSampledAnimation(Animation):
    def __init__(self, indices: List[int], boundary: AnimationBoundary = AnimationBoundary.HOLD):
        super().__init__(boundary)
        self.indices = indices

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        count = len(self.indices)
        if count == 0:
            return -1
        elif count == 1:
            return 0
        else:
            idx = self.map_frame_index(frame_index, self.indices[0], self.indices[-1] + 1)
            return max(bisect.bisect_right(self.indices, idx) - 1, 0)

    def is_dynamic(self, n: int) -> bool:
        return True

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        if not self.indices:
            return True
        else:
            return self.is_frame_index_disabled(frame_index, self.indices[0], self.indices[-1])

    def start_animation_time(self, n: int, fps: float) -> float:
        return self.indices[0] / fps if self.indices else 0.0

    def end_animation_time(self, n: int, fps: float) -> float:
        return (self.indices[-1] + 1) / fps if self.indices else 0.0


class ConstantSpeedAnimation(Animation):
    def __init__(
        self, frames_per_second: float, start_frame: int = 0, boundary: AnimationBoundary = AnimationBoundary.HOLD
    ):
        super().__init__(boundary)
        self.frames_per_second = frames_per_second
        self.start_frame = start_frame

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        if n == 0:
            return -1
        return self.map_frame_index(int(time * self.frames_per_second), 0, n)

    def is_dynamic(self, n: int) -> bool:
        return n > 1 or self.start_frame > 0

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        if n == 0:
            return True
        else:
            return self.is_frame_index_disabled(int(time * self.frames_per_second), 0, n)

    def start_animation_time(self, n: int, fps: float) -> float:
        return self.start_frame / self.frames_per_second

    def end_animation_time(self, n: int, fps: float) -> float:
        return (self.start_frame + n) / self.frames_per_second


class FrameAnimation(Animation):
    def __init__(self, start_frame: int = 0, boundary: AnimationBoundary = AnimationBoundary.HOLD):
        super().__init__(boundary)
        self.start_frame = start_frame

    def get_frame_index(self, n: int, time: float, frame_index: int) -> int:
        if n == 0 or frame_index < self.start_frame:
            return -1
        return self.map_frame_index(frame_index - self.start_frame, 0, n)

    def is_dynamic(self, n: int) -> bool:
        return n > 1 or self.start_frame > 0

    def is_disabled(self, n: int, time: float, frame_index: int) -> bool:
        if n == 0 or frame_index < self.start_frame:
            return True
        else:
            return self.is_frame_index_disabled(frame_index - self.start_frame, 0, n)

    def start_animation_time(self, n: int, frames_per_second: float) -> float:
        return self.start_frame / frames_per_second

    def end_animation_time(self, n: int, frames_per_second: float) -> float:
        return (self.start_frame + n) / frames_per_second


@dataclass
class UploadSettings:
    preupload: bool = True
    batched: bool = True
    async_load: bool = False
    cpu_prefetch_count: int = 0
    gpu_prefetch_count: int = 0


class Property:
    gpu_property: Optional[gpu_property_mod.GpuProperty[Any]]

    def __init__(
        self,
        num_frames: int,
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        self.num_frames = num_frames
        self.name = name
        self.animation = animation if animation is not None else FrameAnimation()
        self.upload = upload if upload is not None else UploadSettings()

        self.current_frame_index = 0 if num_frames > 0 else -1
        self.current_animation_enabled = num_frames > 0

        self.update_callbacks: List[Callable[[Property], None]] = []

    # Public API
    def get_current(self) -> PropertyItem:
        return self.get_frame_by_index(self.current_frame_index)

    def get_frame_by_index_into(self, frame_index: int, out: memoryview, thread_index: int = -1) -> int:
        frame = self.get_frame_by_index(frame_index, thread_index)
        if frame is None:
            return 0
        data = view_bytes(frame)
        size = len(data)
        out[:size] = data
        return size

    def get_frame_by_index(self, frame_index: int, thread_index: int = -1) -> PropertyItem:
        raise NotImplementedError

    def destroy_gpu_property(self) -> None:
        if self.gpu_property is not None:
            self.gpu_property.enqueue_for_destruction()
            self.gpu_property = None

    # Scene API
    def is_dynamic(self) -> bool:
        return self.animation.is_dynamic(self.num_frames)

    def update(self, time: float, frame: int) -> None:
        old_frame = self.current_frame_index
        self.current_frame_index = self.animation.get_frame_index(self.num_frames, time, frame)
        self.current_animation_enabled = not self.animation.is_disabled(self.num_frames, time, frame)

        if old_frame != self.current_frame_index:
            for c in self.update_callbacks:
                c(self)

    def start_animation_time(self, fps: float) -> float:
        return self.animation.start_animation_time(self.num_frames, fps)

    def end_animation_time(self, fps: float) -> float:
        return self.animation.end_animation_time(self.num_frames, fps)

    # Renderer API
    def create(self, r: "Renderer") -> None:
        pass

    def destroy(self) -> None:
        self.destroy_gpu_property()

    # Edit API
    def update_frame(self, frame_index: int, frame: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.UPDATE:
                self.gpu_property.update_frame(frame_index)
            else:
                self.destroy_gpu_property()

    def update_frames(self, frame_indices: List[int], frames: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.UPDATE:
                self.gpu_property.update_frames(frame_indices)
            else:
                self.destroy_gpu_property()

    def update_frame_range(self, start: int, frames: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.UPDATE:
                self.gpu_property.update_frame_range(start, start + len(frames))
            else:
                self.destroy_gpu_property()

    def append_frame(self, frame: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.APPEND:
                self.gpu_property.append_frame()
            else:
                self.destroy_gpu_property()

    def append_frames(self, frames: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.APPEND:
                self.gpu_property.append_frames(len(frames))
            else:
                self.destroy_gpu_property()

    def pop_frame(self) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.POP:
                self.gpu_property.pop_frame()
            else:
                self.destroy_gpu_property()

    def pop_frames(self, count: int) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.POP:
                self.gpu_property.pop_frames(count)
            else:
                self.destroy_gpu_property()

    def insert_frame(self, frame_index: int, frame: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.INSERT:
                self.gpu_property.insert_frame(frame_index)
            else:
                self.destroy_gpu_property()

    def insert_frames(self, index: int, frames: Any) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.INSERT:
                self.gpu_property.insert_frames(index, len(frames))
            else:
                self.destroy_gpu_property()

    def remove_frame(self, frame_index: int) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.REMOVE:
                self.gpu_property.remove_frame(frame_index)
            else:
                self.destroy_gpu_property()

    def remove_frames(self, frame_indices: List[int]) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.REMOVE:
                self.gpu_property.remove_frames(frame_indices)
            else:
                self.destroy_gpu_property()

    def remove_frame_range(self, start: int, stop: int) -> None:
        if self.gpu_property is not None:
            if self.gpu_property.supported_operations & gpu_property_mod.GpuPropertySupportedOperations.REMOVE:
                self.gpu_property.remove_frame_range(start, stop)
            else:
                self.destroy_gpu_property()


class BufferProperty(Property):
    def __init__(
        self,
        max_size: int,
        num_frames: int,
        dtype: DTypeLike,
        shape: Tuple[int, ...],
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        super().__init__(num_frames, animation, upload, name)
        self.max_size = max_size
        self.dtype = dtype
        self.shape = shape

        self.gpu_property: Optional[gpu_property_mod.GpuProperty[gpu_property_mod.GpuBufferView]] = None
        self.gpu_usage = BufferUsageFlags(0)
        self.gpu_stage = PipelineStageFlags(0)

    # Public API
    def use_gpu(self, usage: BufferUsageFlags, stage: PipelineStageFlags) -> "BufferProperty":
        self.gpu_usage |= usage
        self.gpu_stage |= stage
        return self

    def get_current_gpu(self) -> gpu_property_mod.GpuBufferView:
        assert self.gpu_property is not None
        return self.gpu_property.get_current()

    # Renderer API
    def create(self, r: "Renderer") -> None:
        if self.gpu_usage and self.gpu_property is None:
            # HACK: disable shader device address and acceleration structure usages if not supported.
            if not (r.device.features & DeviceFeatures.RAY_QUERY):
                self.gpu_usage &= ~(
                    BufferUsageFlags.SHADER_DEVICE_ADDRESS | BufferUsageFlags.ACCELERATION_STRUCTURE_INPUT
                )

            if self.upload.preupload:
                self.gpu_property = gpu_property_mod.GpuBufferPreuploadedProperty(r, self, self.name)
            else:
                self.gpu_property = gpu_property_mod.GpuBufferStreamingProperty(r, self, self.name)


class ImageProperty(Property):
    def __init__(
        self,
        width: int,
        height: int,
        format: Format,
        num_frames: int,
        depth: Optional[int] = None,
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        super().__init__(num_frames, animation, upload, name)
        self.width = width
        self.height = height
        self.depth = depth
        self.format = format

        self.gpu_property: Optional[gpu_property_mod.GpuProperty[gpu_property_mod.GpuImageView]] = None
        self.gpu_usage = ImageUsageFlags(0)
        self.gpu_stage = PipelineStageFlags(0)
        self.gpu_layout = ImageLayout.UNDEFINED
        self.gpu_srgb = False
        self.gpu_mips = False

    # Public API
    def use_gpu(
        self,
        usage: ImageUsageFlags,
        layout: ImageLayout,
        stage: PipelineStageFlags,
        srgb: bool = False,
        mips: bool = False,
    ) -> "ImageProperty":
        if self.gpu_layout == ImageLayout.UNDEFINED:
            self.gpu_layout = layout
        elif self.gpu_layout != layout:
            # Fallback to general if we need multiple layouts
            self.gpu_layout = ImageLayout.GENERAL
        self.gpu_usage |= usage
        self.gpu_stage |= stage
        self.gpu_srgb |= srgb
        self.gpu_mips |= mips
        return self

    def get_current_gpu(self) -> gpu_property_mod.GpuImageView:
        assert self.gpu_property is not None
        return self.gpu_property.get_current()

    # Renderer API
    def create(self, r: "Renderer") -> None:
        if self.gpu_usage and self.gpu_property is None:
            if self.gpu_mips and self.depth is not None:
                raise RuntimeError("mips are not supported on 3D ImageProperty")

            if self.upload.preupload:
                self.gpu_property = gpu_property_mod.GpuImagePreuploadedProperty(r, self, self.name)
            else:
                self.gpu_property = gpu_property_mod.GpuImageStreamingProperty(r, self, self.name)


class ShapeError(Exception):
    def __init__(
        self,
        expected: Optional[Tuple[int, ...]],
        got: Optional[Tuple[int, ...]],
    ):
        self.expected = expected
        self.got = got

    def __str__(self) -> str:
        return f"Shape mismatch. Expected: {self.expected}. Got: {self.got}"


def shape_match(
    expected_shape: Optional[Tuple[int, ...]],
    got_shape: Optional[Tuple[int, ...]],
) -> bool:
    if expected_shape is None:
        return got_shape is None
    else:
        if got_shape is None:
            return False

    if len(expected_shape) != len(got_shape):
        return False
    for a, b in zip(expected_shape, got_shape):
        if a != -1 and a != b:
            return False
    return True


class ArrayBufferProperty(BufferProperty):
    def __init__(
        self,
        data: ArrayLike,
        dtype: Optional[DTypeLike] = None,
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        data = np.atleast_1d(np.asarray(data, dtype, order="C"))
        max_size = data.itemsize * np.prod(data.shape[1:], dtype=np.int64)
        super().__init__(int(max_size), len(data), data.dtype, data.shape[1:], animation, upload, name)
        self.data = data

    # Public API
    def get_frame_by_index(self, frame_index: int, thread_index: int = -1) -> PropertyItem:
        return self.data[frame_index]  # type: ignore

    # Renderer API
    def create(self, r: "Renderer") -> None:
        if self.gpu_usage and self.gpu_property is None:
            # HACK: disable shader device address and acceleration structure usages if not supported.
            if not (r.device.features & DeviceFeatures.RAY_QUERY):
                self.gpu_usage &= ~(
                    BufferUsageFlags.SHADER_DEVICE_ADDRESS | BufferUsageFlags.ACCELERATION_STRUCTURE_INPUT
                )

            if self.upload.preupload:
                if self.upload.batched:
                    self.gpu_property = gpu_property_mod.GpuBufferPreuploadedArrayProperty(
                        self.data, r, self, self.name
                    )
                else:
                    self.gpu_property = gpu_property_mod.GpuBufferPreuploadedProperty(r, self, self.name)
            else:
                self.gpu_property = gpu_property_mod.GpuBufferStreamingProperty(r, self, self.name)

    # Edit API
    def update_frame(self, frame_index: int, frame: ArrayLike) -> None:
        self.data[frame_index] = frame
        super().update_frame(frame_index, frame)

    def update_frames(self, frame_indices: List[int], frames: ArrayLike) -> None:
        self.data[frame_indices] = frames
        super().update_frames(frame_indices, frames)

    def update_frame_range(self, start: int, frames: NDArray[Any]) -> None:
        self.data[start : start + len(frames)] = frames
        super().update_frame_range(start, frames)

    def append_frame(self, frame: ArrayLike) -> None:
        self.data = np.append(self.data, np.asarray(frame)[np.newaxis], axis=0)
        self.num_frames = self.data.shape[0]
        super().append_frame(frame)

    def append_frames(self, frames: ArrayLike) -> None:
        self.data = np.append(self.data, frames, axis=0)
        self.num_frames = self.data.shape[0]
        super().append_frames(frames)

    def pop_frame(self) -> None:
        self.remove_frame(self.num_frames - 1)

    def pop_frames(self, count: int) -> None:
        self.remove_frame_range(self.num_frames - count, self.num_frames)

    def insert_frame(self, frame_index: int, frame: ArrayLike) -> None:
        self.data = np.insert(self.data, frame_index, frame, axis=0)
        self.num_frames = self.data.shape[0]
        super().insert_frame(frame_index, frame)

    def insert_frames(self, index: int, frames: ArrayLike) -> None:
        self.data = np.insert(self.data, index, frames, axis=0)
        self.num_frames = self.data.shape[0]
        super().insert_frames(index, frames)

    def remove_frame(self, frame_index: int) -> None:
        self.data = np.delete(self.data, frame_index, axis=0)
        self.num_frames = self.data.shape[0]
        super().remove_frame(frame_index)

    def remove_frames(self, frame_indices: List[int]) -> None:
        self.data = np.delete(self.data, frame_indices, axis=0)
        self.num_frames = self.data.shape[0]
        super().remove_frames(frame_indices)

    def remove_frame_range(self, start: int, stop: int) -> None:
        self.data = np.delete(self.data, slice(start, stop), axis=0)
        self.num_frames = self.data.shape[0]
        super().remove_frame_range(start, stop)


class ListBufferProperty(BufferProperty):
    def __init__(
        self,
        data: List[ArrayLike],
        dtype: Optional[DTypeLike] = None,
        shape: Tuple[int, ...] = (),
        max_size: int = 0,
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        property_data = []
        for i in range(len(data)):
            a = np.asarray(data[i], dtype, order="C")

            # Shape checking
            if not shape_match(shape, a.shape):
                raise ShapeError(shape, a.shape)

            # If no dtype was given, use dtype inferred for first element
            if dtype is None:
                dtype = a.dtype

            property_data.append(a)
            max_size = max(max_size, a.nbytes)

        super().__init__(max_size, len(property_data), dtype, shape, animation, upload, name)
        self.data = property_data

    # Public API
    def get_frame_by_index(self, frame_index: int, thread_index: int = -1) -> PropertyItem:
        return self.data[frame_index]

    # Edit API
    def update_frame(self, frame_index: int, frame: NDArray[Any]) -> None:
        self.data[frame_index] = frame
        super().update_frame(frame_index, frame)

    def update_frames(self, frame_indices: List[int], frames: List[NDArray[Any]]) -> None:
        for i, f in zip(frame_indices, frames):
            self.data[i] = f
        super().update_frames(frame_indices, frames)

    def update_frame_range(self, start: int, frames: List[NDArray[Any]]) -> None:
        self.data[start : len(frames)] = frames
        super().update_frame_range(start, frames)

    def append_frame(self, frame: NDArray[Any]) -> None:
        self.data.append(frame)
        self.num_frames = len(self.data)
        super().append_frame(frame)

    def append_frames(self, frames: List[NDArray[Any]]) -> None:
        self.data.extend(frames)
        self.num_frames = len(self.data)
        super().append_frames(frames)

    def pop_frame(self) -> None:
        self.data.pop()
        self.num_frames = len(self.data)
        super().pop_frame()

    def pop_frames(self, count: int) -> None:
        if count > 0:
            del self.data[-count:]
        self.num_frames = len(self.data)
        super().pop_frames(count)

    def insert_frame(self, frame_index: int, frame: NDArray[Any]) -> None:
        self.data.insert(frame_index, frame)
        self.num_frames = len(self.data)
        super().insert_frame(frame_index, frame)

    def insert_frames(self, index: int, frames: List[NDArray[Any]]) -> None:
        self.data = self.data[:index] + frames + self.data[index:]
        self.num_frames = len(self.data)
        super().insert_frames(index, frames)

    def remove_frame(self, frame_index: int) -> None:
        self.data.pop(frame_index)
        self.num_frames = len(self.data)
        super().remove_frame(frame_index)

    def remove_frames(self, frame_indices: List[int]) -> None:
        frame_indices_set = set(frame_indices)
        self.data = [d for i, d in enumerate(self.data) if i not in frame_indices_set]
        self.num_frames = len(self.data)
        super().remove_frames(frame_indices)

    def remove_frame_range(self, start: int, stop: int) -> None:
        self.data = self.data[:start] + self.data[stop:]
        self.num_frames = len(self.data)
        super().remove_frame_range(start, stop)


class ArrayImageProperty(ImageProperty):
    def __init__(
        self,
        in_data: ArrayLike,
        format: Optional[Format] = None,
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        data = np.asarray(in_data, order="C")

        if len(data.shape) == 4:
            height, width, channels = data.shape[1:4]
            depth = None
            if format is None:
                format = format_from_channels_dtype(channels, data.dtype)
            else:
                # TODO: check that format is compatible with data (maybe allow some conversions?)
                pass
        elif len(data.shape) == 5:
            depth, height, width, channels = data.shape[1:5]
            if format is None:
                format = format_from_channels_dtype(channels, data.dtype)
        else:
            raise ShapeError((-1, -1, -1), data.shape)

        super().__init__(width, height, format, len(data), depth, animation, upload, name)
        self.data = data

    # Public API
    def get_frame_by_index(self, frame_index: int, thread_index: int = -1) -> PropertyItem:
        return self.data[frame_index]  # type: ignore

    # Edit API
    def update_frame(self, frame_index: int, frame: ArrayLike) -> None:
        self.data[frame_index] = frame
        super().update_frame(frame_index, frame)

    def update_frames(self, frame_indices: List[int], frames: ArrayLike) -> None:
        self.data[frame_indices] = frames
        super().update_frames(frame_indices, frames)

    def update_frame_range(self, start: int, frames: NDArray[Any]) -> None:
        self.data[start : start + len(frames)] = frames
        super().update_frame_range(start, frames)

    def append_frame(self, frame: ArrayLike) -> None:
        self.data = np.append(self.data, np.asarray(frame)[np.newaxis], axis=0)
        self.num_frames = self.data.shape[0]
        super().append_frame(frame)

    def append_frames(self, frames: ArrayLike) -> None:
        self.data = np.append(self.data, frames, axis=0)
        self.num_frames = self.data.shape[0]
        super().append_frames(frames)

    def pop_frame(self) -> None:
        self.remove_frame(self.num_frames - 1)

    def pop_frames(self, count: int) -> None:
        self.remove_frame_range(self.num_frames - count, self.num_frames)

    def insert_frame(self, frame_index: int, frame: ArrayLike) -> None:
        self.data = np.insert(self.data, frame_index, frame, axis=0)
        self.num_frames = self.data.shape[0]
        super().insert_frame(frame_index, frame)

    def insert_frames(self, index: int, frames: ArrayLike) -> None:
        self.data = np.insert(self.data, index, frames, axis=0)
        self.num_frames = self.data.shape[0]
        super().insert_frames(index, frames)

    def remove_frame(self, frame_index: int) -> None:
        self.data = np.delete(self.data, frame_index, axis=0)
        self.num_frames = self.data.shape[0]
        super().remove_frame(frame_index)

    def remove_frames(self, frame_indices: List[int]) -> None:
        self.data = np.delete(self.data, frame_indices, axis=0)
        self.num_frames = self.data.shape[0]
        super().remove_frames(frame_indices)

    def remove_frame_range(self, start: int, stop: int) -> None:
        self.data = np.delete(self.data, slice(start, stop), axis=0)
        self.num_frames = self.data.shape[0]
        super().remove_frame_range(start, stop)


class ListImageProperty(ImageProperty):
    def __init__(
        self,
        data: List[ArrayLike],
        format: Optional[Format] = None,
        animation: Optional[Animation] = None,
        upload: Optional[UploadSettings] = None,
        name: str = "",
    ):
        property_data: List[NDArray[Any]] = []
        property_shape: Optional[Tuple[int, int, int]] = None
        property_dtype: DTypeLike = None
        for i in range(len(data)):
            a = np.asarray(data[i], property_dtype, order="C")
            if len(a.shape) != 3 or len(a.shape) != 4:
                raise ShapeError((-1, -1, -1), a.shape)
            if property_shape is None:
                property_shape = a.shape
            else:
                if property_shape != a.shape:
                    raise RuntimeError(
                        f"ListImageProperty data elements must all have the same shape. First has {property_shape}. Element {i} has {a.shape}"
                    )
            if property_dtype is None:
                property_dtype = a.dtype
            property_data.append(a)

        if len(property_data[0].shape) == 3:
            height, width, channels = property_data[0].shape
            depth = None
            if format is None:
                format = format_from_channels_dtype(channels, property_data[0].dtype)
            else:
                # TODO: check that format is compatible with data (maybe allow some conversions?)
                pass
        else:
            assert len(property_data[0].shape) == 4
            depth, height, width, channels = property_data[0].shape
            if format is None:
                format = format_from_channels_dtype(channels, property_data[0].dtype)

        super().__init__(width, height, format, len(property_data), depth, animation, upload, name)
        self.data = property_data

    # Public API
    def get_frame_by_index(self, frame_index: int, thread_index: int = -1) -> PropertyItem:
        return self.data[frame_index]

    # Edit API
    def update_frame(self, frame_index: int, frame: NDArray[Any]) -> None:
        self.data[frame_index] = frame
        super().update_frame(frame_index, frame)

    def update_frames(self, frame_indices: List[int], frames: List[NDArray[Any]]) -> None:
        for i, f in zip(frame_indices, frames):
            self.data[i] = f
        super().update_frames(frame_indices, frames)

    def update_frame_range(self, start: int, frames: List[NDArray[Any]]) -> None:
        self.data[start : len(frames)] = frames
        super().update_frame_range(start, frames)

    def append_frame(self, frame: NDArray[Any]) -> None:
        self.data.append(frame)
        self.num_frames = len(self.data)
        super().append_frame(frame)

    def append_frames(self, frames: List[NDArray[Any]]) -> None:
        self.data.extend(frames)
        self.num_frames = len(self.data)
        super().append_frames(frames)

    def pop_frame(self) -> None:
        self.data.pop()
        self.num_frames = len(self.data)
        super().pop_frame()

    def pop_frames(self, count: int) -> None:
        if count > 0:
            del self.data[-count:]
        self.num_frames = len(self.data)
        super().pop_frames(count)

    def insert_frame(self, frame_index: int, frame: NDArray[Any]) -> None:
        self.data.insert(frame_index, frame)
        self.num_frames = len(self.data)
        super().insert_frame(frame_index, frame)

    def insert_frames(self, index: int, frames: List[NDArray[Any]]) -> None:
        self.data = self.data[:index] + frames + self.data[index:]
        self.num_frames = len(self.data)
        super().insert_frames(index, frames)

    def remove_frame(self, frame_index: int) -> None:
        self.data.pop(frame_index)
        self.num_frames = len(self.data)
        super().remove_frame(frame_index)

    def remove_frames(self, frame_indices: List[int]) -> None:
        frame_indices_set = set(frame_indices)
        self.data = [d for i, d in enumerate(self.data) if i not in frame_indices_set]
        self.num_frames = len(self.data)
        super().remove_frames(frame_indices)

    def remove_frame_range(self, start: int, stop: int) -> None:
        self.data = self.data[:start] + self.data[stop:]
        self.num_frames = len(self.data)
        super().remove_frame_range(start, stop)


def as_buffer_property(
    value: Union[BufferProperty, List[ArrayLike], ArrayLike],
    dtype: Optional[DTypeLike] = None,
    shape: Tuple[int, ...] = (),
    name: str = "",
) -> BufferProperty:
    if isinstance(value, BufferProperty):
        if not shape_match(shape, value.shape):
            raise ShapeError(shape, value.shape)
        if dtype is not None and not np.dtype(dtype) == value.dtype:
            raise TypeError(f"dtype mismatch. Expected: {dtype}. Got: {value.dtype}")
        if not value.name:
            value.name = name
        return value
    elif isinstance(value, List):
        return ListBufferProperty(value, dtype, shape, name=name)
    else:
        value = np.atleast_1d(np.asarray(value, dtype, order="C"))
        if shape is None:
            if len(value.shape) > 1:
                raise ShapeError((1,), value.shape[1:])
        else:
            if shape_match(shape, value.shape):
                # Implicitly add animation dimension
                value = value[np.newaxis]
            elif shape_match(shape, value.shape[1:]):
                # Shape already matches
                pass
            else:
                raise ShapeError(shape, value.shape)
        return ArrayBufferProperty(value, dtype, name=name)


def as_image_property(
    value: Union[ImageProperty, List[ArrayLike], ArrayLike],
    is_3d: bool = False,
    name: str = "",
) -> ImageProperty:
    if isinstance(value, ImageProperty):
        if not value.name:
            value.name = name
        return value
    elif isinstance(value, List):
        return ListImageProperty(value, None, name=name)
    else:
        data = np.asarray(value)
        if is_3d:
            dims = 3
        else:
            dims = 2

        if data.ndim == dims:
            # Implicitly add animation and channel dimension
            data = data[np.newaxis, ..., np.newaxis]
        elif data.ndim == dims + 1:
            # Implicitly add animation dimension
            data = data[np.newaxis]
        elif data.ndim == dims + 2:
            # Shape already matches
            pass
        else:
            raise ShapeError((-1, -1, -1), data.shape)
        return ArrayImageProperty(data, None, name=name)
